from setuptools import setup


setup(name='roseng',
      version='0.3',
      description='ROS Engine',
      author='Dirk Zhou',
      author_email='dontbmh2011@gmail.com',
      license='MIT',
      packages=['roseng'],
      python_requires='>=3.6',
      zip_safe=False)
